/* Core Hardware: Common macros
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/** \brief CPWAIT macro as defined in the XScale Core Manual on Page 2.3.3-31
 * \param reg register to use (will contain the translation table base register
 *  afterwards, but this should be ignored
 * \note look at the XScale Core Manual on Page 2.3.3-31 */
.macro CPWAIT reg
	/* load translation table base register */
	mrc p15, 0, \reg, c2, c0, 0
	/* nop */
	mov \reg, \reg
	/* next instruction */
	sub pc, pc, #0x4
.endm

/** \brief Set Bits in the Control Register (Co-Processor Register 15)
 * \param reg register to use (will contain the new CP15 value afterwards)
 * \param mask bits to be set in the register. Must match the constraints of
 *  the orr-mnemonic
 * \note look at the XScale Core Manual on Page 7.2.2-83 (Control Register) */
.macro SET_CONTROL_BITS reg, mask
	/* load old value */
	mrc p15, 0, \reg, c1, c0, 0
	/* set bits */
	orr \reg, \reg, \mask
	/* store back new value */
	mcr p15, 0, \reg, c1, c0, 0
.endm

/** \brief Unset Bits in the Control Register (Co-Processor Register 15)
 * \param reg register to use (will contain the new CP15 value afterwards)
 * \param mask bits to be unset in the register. Must match the constraints of
 *  the bic-mnemonic
 * \note look at the XScale Core Manual on Page 7.2.2-83 (Control Register) */
.macro UNSET_CONTROL_BITS reg, mask
	/* load old value */
	mrc p15, 0, \reg, c1, c0, 0
	/* clear bits */
	bic \reg, \reg, \mask
	/* store back new value */
	mcr p15, 0, \reg, c1, c0, 0
.endm

/** \brief Test a Bit in the Control Register (Co-Processor Register 15)
 * \param reg register to use
 * \param bit bit to be tested, e.g. #(1 << 10)
 * \return (in reg) 0x1 (\c TRUE) iff the bit is set, 0x0 otherwise
 * \note look at the XScale Core Manual on Page 7.2.2-83 (Control Register) */
.macro TEST_CONTROL_BITS reg, bit
	/* read control register into reg */
	mrc p15, 0, \reg, c1, c0, 0
	/* test reg & bit */
	tst \reg, \bit
	/* set reg=0 if the bit is not set */
	moveq r0, #0x0
	/* set reg=1 if the bit is set */
	movne r0, #0x1
.endm

/** Control Register Bit definitions
 * \note look at the XScale Core Manual on Page 7.2.2-83 (Control Register) */

/** Instruction Cache Enable Bit */
CP15_CONTROL_I = (1 << 12)
/** Branch Target Buffer Enable Bit */
CP15_CONTROL_Z = (1 << 11)
/** Data Cache Enable Bit */
CP15_CONTROL_C = (1 << 2)
/** Memory Management Unit Enable Bit */
CP15_CONTROL_M = (1 << 0)


/** \brief Invalidate Instruction Cache
 * \note as a side-effect, this will also invalidate the btb
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro ICACHE_INVALIDATE
	mcr p15, 0, r0, c7, c5, 0
.endm

/** \brief Invalidate Instruction Cache Line
 * \param line line to be invalidated
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro ICACHE_INVALIDATE_LINE line
	mcr p15, 0, \line, c7, c5, 1
.endm

/** \brief Invalidate Data Cache
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro DCACHE_INVALIDATE
	mcr p15, 0, r0, c7, c6, 0
.endm

/** \brief Invalidate Data Cache Line
 * \param line line to be invalidated
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro DCACHE_INVALIDATE_LINE line
	mcr p15, 0, \line, c7, c6, 1
.endm

/** \brief Clean Data Cache Line
 * \param line line to be cleaned
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro DCACHE_CLEAN_LINE line
	mcr p15, 0, \line, c7, c10, 1
.endm

/** \brief Drain Write Buffer
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro DCACHE_DRAIN_BUFFER
	mcr p15, 0, r0, c7, c10, 4
.endm

/** \brief Invalidate Branch Target Buffer
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro BTB_INVALIDATE
	mcr p15, 0, r0, c7, c5, 6
.endm

/** \brief allocate a line in the data cache
 * \param addr address of line to be allocated
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
.macro DCACHE_ALLOCATE addr
	MCR P15, 0, \addr, c7, c2, 5
.endm



/** \brief Invalidate Instruction and Data TLB
 * \note look at the XScale Core Manual on Page 7.2.9-87 (TLB Operations) */
.macro TLB_INVALIDATE
	mcr p15, 0, r0, c8, c7, 0
.endm

/** \brief Invalidate Instruction TLB
 * \note look at the XScale Core Manual on Page 7.2.9-87 (TLB Operations) */
.macro ITLB_INVALIDATE
	mcr p15, 0, r0, c8, c5, 0
.endm

/** \brief Invalidate Data TLB
 * \note look at the XScale Core Manual on Page 7.2.9-87 (TLB Operations) */
.macro DTLB_INVALIDATE
	mcr p15, 0, r0, c8, c6, 0
.endm



/** \brief Unlock Instruction Cache
 * \note look at the XScale Core Manual on Page 7.2.10-90 (Cache Lock Down) */
.macro ICACHE_UNLOCK
	mcr p15, 0, r0, c9, c1, 1
.endm

/** \brief Unlock Data Cache
 * \note look at the XScale Core Manual on Page 7.2.10-90 (Cache Lock Down) */
.macro DCACHE_UNLOCK
	mcr p15, 0, r0, c9, c2, 1
.endm


/** \brief Unlock Instruction TLB
 * \note look at the XScale Core Manual on Page 7.2.11-91 (TLB Lock Down) */
.macro ITLB_UNLOCK
	mcr p15, 0, r0, c10, c4, 1
.endm

/** \brief Unlock Data TLB
 * \note look at the XScale Core Manual on Page 7.2.11-91 (TLB Lock Down) */
.macro DTLB_UNLOCK
	mcr p15, 0, r0, c10, c8, 1
.endm

