/* CFI Flash driver
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern,

This file is part of the Imote2 Bootloader.

It contains ideas taken from the blob project and the linux mtd device driver,
both licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stddef.h>
#include <malloc.h>

#include <inttypes.h>
#include <os_arm_bare_registers_def.h>
#include "dmalloc.h"
#include "flash_internal.h"
#include "flash.h"
#include "debug.h"
#include "config.h"

/** \defgroup flash Flash Subsystem
 * @{
 */

/** \file flash.c
 * \brief routines for manipulating the CFI flash device
 */

static void init(struct Driver * driver);

/** \brief Flash Driver structure */
struct Driver flashDriver = {
	.name = "Flash",
	
	.init = &init
};

/** \brief write a halfword to the chip
 * \param block block base address
 * \param offset offset in the block
 * \param data half word to be written */
static inline void write(uint32_t block, uint32_t offset,
	uint16_t data) {
	*ADDR16(block, offset) = data;
}

/** \brief read a halfword from the chip
 * \param block block base address
 * \param offset offset in the block
 * \return data */
static inline uint16_t read(uint32_t block, uint32_t offset) {
	return *ADDR16(block, offset);
}

/** \brief put device in <a>READ ARRAY</a> mode
 * \param block block base address */
static inline void cmdReadArray(uint32_t block) {
	write(block, 0x0, CMD_READ_ARRAY);
}

/** \brief put device in <a>READ QUERY</a> mode
 * \param block block base address */
static inline void cmdReadQuery(uint32_t block) {
	write(block, 0x0, CMD_READ_QUERY);
}

/** \brief clear device status */
static inline void cmdClearStatus() {
	write(0x0, 0x0, CMD_CLEAR_STATUS);
}

/** \brief setup erase mode
 * \param block block base address */
static inline void cmdEraseSetup(uint32_t block) {
	write(block, 0x0, CMD_BLOCK_ERASE);
}

/** \brief setup program mode
 * \param block block base address */
static inline void cmdProgramSetup(uint32_t block) {
	write(block, 0x0, CMD_WORD_PROGRAM);
}

/** \brief setup program mode
 * \param block block base address */
static inline void cmdProgramBufferSetup(uint32_t block) {
	write(block, 0x0, CMD_WRITE_TO_BUFFER_SETUP);
}

/** \brief setup lock mode
 * \param block block base address */
static inline void cmdBlockLock(uint32_t block) {
	write(block, 0x0, CMD_LOCK_BLOCK);
}

/** \brief read status from chip.
 * \warning needs the device to be in <a>READ STATUS</a> mode
 * \return status */
static inline uint16_t readStatus(uint32_t block) {
	return read(block, 0x0);
}

/** read status from chip until the chip is ready
 *
 * reads the status until the ready bit is set and resets device
 *  into read array mode.
 *
 * \warning needs the device to be in <a>READ STATUS</a> mode
 * \return status */
static inline uint8_t readStatusWait(uint32_t block) {
	uint8_t status;
	
	/* spin while not ready */
	do {
		status = read(block, 0x0);
	} while ((status & STATUS_READY) == 0);
	
	/* return to READ ARRAY mode */
	cmdReadArray(block);
	
	/* return status */
	return status;
}

#ifdef DEBUG_FLASH
/** \brief allocates a ::CfiIdent structure, fills and returns it
 *
 * \return ident structure or \c NULL iff allocation failed */ 
static struct CfiIdent * queryChip() {
	struct CfiIdent * ident;
	uint8_t num_erase_blocks;
	int i;
	
	/* put flash in read query mode */
	cmdReadQuery(0x0);
	/* read number of erase blocks */
	num_erase_blocks = read(0x0, OFFSET_CFI_IDENT +
		offsetof(struct CfiIdent, numEraseRegions));
	/* put device into read array mode as dmalloc is called */
	cmdReadArray(0x0);
	
	/* allocate size of basic structure plus the size for the erase blocks */
	ident = (struct CfiIdent*)dmalloc(sizeof(struct CfiIdent) +
		4 * num_erase_blocks);
	if (ident == NULL)
		return NULL;
	
	/* put flash in read query mode again */
	cmdReadQuery(0x0);
	/* read table */
	for (i = 0; i < sizeof(struct CfiIdent) + 4 * num_erase_blocks; ++i)
		((uint8_t*)ident)[i] = read(0x0, OFFSET_CFI_IDENT + i);
	/* return to read array mode again */
	cmdReadArray(0x0);
	
	return ident;
}

/** \brief allocates an ::CfiPriIntelExt structure,
 *   fills and returns it
 *
 * \param offset Algorithm Address as returned by queryChip()->extTable
 * \return intel ident structure or \c NULL iff allocation failed */
static struct CfiPriIntelExt * queryIntelPRI(uint16_t offset) {
	struct CfiPriIntelExt * pri;
	int i;
	
	/* allocate structure */
	pri = (struct CfiPriIntelExt*)dmalloc(sizeof(struct CfiPriIntelExt));
	if (pri == NULL)
		return NULL;
	
	/* put flash in read query mode */
	cmdReadQuery(0x0);
	/* read table */
	for (i = 0; i < sizeof(struct CfiPriIntelExt); ++i)
		((uint8_t*)pri)[i] = read(0x0, offset + i);
	/* return to read array mode */
	cmdReadArray(0x0);
	
	return pri;
}

/** \brief reads the flash info using queryChip() and queryIntelPRI()
 * and prints some fancy information about the chip */
static void flashInfo() {
	struct CfiIdent * ident;
	struct CfiPriIntelExt * pri;
	int i;
	
	/* read CFI info structure */
	ident = queryChip();
	
	if (ident == NULL) {
		debugf("Could not allocate space for flash info");
		return;
	}
	
	/* check query */
	if (ident->qry[0] != 'Q' || ident->qry[1] != 'R' || ident->qry[2] != 'Y') {
		debugf("Invalid CFI ident structure.\n");
		dfree(ident);
		return;
	}
	
	/* print query information */
	debugf("Primary Vendor Command Set: %4.4X", ident->cmdSet);
	if (ident->extTable)
		debugf("Primary Algorithm Table at %4.4X", ident->extTable);
	else
		debugf("No Primary Algorithm Table");

	debugf("Alternative Vendor Command Set: %4.4X", ident->altCmdSet);
	if (ident->altExtTable)
		debugf("Alternate Algorithm Table at %4.4X", ident->altExtTable);
	else
		debugf("No Alternate Algorithm Table");


	debugf("Vcc Minimum: %2d.%d V", ident->vccMin >> 4,
		ident->vccMin & 0xf);
	debugf("Vcc Maximum: %2d.%d V", ident->vccMax >> 4,
		ident->vccMax & 0xf);
	if (ident->vppMin) {
		debugf("Vpp Minimum: %2d.%d V", ident->vppMin >> 4,
			ident->vppMin & 0xf);
		debugf("Vpp Maximum: %2d.%d V", ident->vppMax >> 4,
			ident->vppMax & 0xf);
	} else
		debugf("No Vpp line");

	debugf("Typical byte/word write timeout: %d µs",
		1 << ident->wordWriteTimeoutTyp);
	debugf("Maximum byte/word write timeout: %d µs",
		(1 << ident->wordWriteTimeoutMax) *
		(1 << ident->wordWriteTimeoutTyp));

	if (ident->bufWriteTimeoutTyp || ident->bufWriteTimeoutMax) {
		debugf("Typical full buffer write timeout: %d µs",
			1 << ident->bufWriteTimeoutTyp);
		debugf("Maximum full buffer write timeout: %d µs",
			(1 << ident->bufWriteTimeoutMax) *
			(1 << ident->bufWriteTimeoutTyp));
	} else
		debugf("Full buffer write not supported");

	debugf("Typical block erase timeout: %d ms",
		1 << ident->blockEraseTimeoutTyp);
	debugf("Maximum block erase timeout: %d ms",
		(1 << ident->blockEraseTimeoutMax) *
		(1 << ident->blockEraseTimeoutTyp));
	if (ident->chipEraseTimeoutTyp || ident->chipEraseTimeoutMax) {
		debugf("Typical chip erase timeout: %d ms",
			1 << ident->chipEraseTimeoutTyp);
		debugf("Maximum chip erase timeout: %d ms",
			(1 << ident->chipEraseTimeoutMax) *
			(1 << ident->chipEraseTimeoutTyp));
	} else
		debugf("Chip erase not supported");

	debugf("Device size: 0x%X bytes (%d MiB)",
		1 << ident->devSize, 1 << (ident->devSize - 20));
	debugf("Flash Device Interface description: 0x%4.4X",
		ident->interfaceDesc);

	debugf("Max. bytes in buffer write: 0x%x", 1 << ident->maxBufWriteSize);
	debugf("Number of Erase Block Regions: %d", ident->numEraseRegions);
	
	for (i = 0; i < ident->numEraseRegions; ++i) {
		debugf("  Erase Region #%d: BlockSize 0x%4.4X bytes, %d blocks\n",
			i, (ident->eraseRegionInfo[i] >> 8) & ~0xff,
			(ident->eraseRegionInfo[i] & 0xffff) + 1);
	}
	
	/* read intel ext info structure */
	pri = queryIntelPRI(ident->extTable);
	
	/* free CFI info structure */
	dfree(ident);
	
	if (pri == NULL) {
		debugf("Could not allocate space for extended flash info");
		return;
	}
	
	/* check pri query */
	if (pri->pri[0] != 'P' || pri->pri[1] != 'R' || pri->pri[2] != 'I') {
		debugf("Invalid PRI ident structure.");
		dfree(pri);
		return;
	}
	
	/* print information */
	debugf("Extended Query version %c.%c", pri->majorVersion,
		pri->minorVersion);
	debugf("Feature/Command Support:      %4.4X", pri->featureSupport);
	debugf("    - Chip Erase:              %s",
		pri->featureSupport & 1 ? "supported" : "unsupported");
	debugf("    - Suspend Erase:           %s",
		pri->featureSupport & 2 ? "supported" : "unsupported");
	debugf("    - Suspend Program:         %s",
		pri->featureSupport & 4 ? "supported" : "unsupported");
	debugf("    - Legacy Lock/Unlock:      %s",
		pri->featureSupport & 8 ? "supported" : "unsupported");
	debugf("    - Queued Erase:            %s",
		pri->featureSupport & 16 ? "supported" : "unsupported");
	debugf("    - Instant block lock:      %s",
		pri->featureSupport & 32 ? "supported" : "unsupported");
	debugf("    - Protection Bits:         %s",
		pri->featureSupport & 64 ? "supported" : "unsupported");
	debugf("    - Page-mode read:          %s",
		pri->featureSupport & 128 ? "supported" : "unsupported");
	debugf("    - Synchronous read:        %s",
		pri->featureSupport & 256 ? "supported" : "unsupported");
	debugf("    - Simultaneous operations: %s",
		pri->featureSupport & 512 ? "supported" : "unsupported");
	debugf("    - Extended Flash Array:    %s",
		pri->featureSupport & 1024 ? "supported" : "unsupported");
	for (i = 11; i < 32; i++) {
		if (pri->featureSupport & (1 << i))
			debugf("    - Unknown Bit %X:      supported", i);
	}

	debugf("Supported functions after Suspend: %2.2X",
		pri->suspendCmdSupport);
	debugf("    - Program after Erase Suspend: %s",
		pri->suspendCmdSupport & 1 ? "supported" : "unsupported");
	for (i = 1; i < 8; i++) {
		if (pri->suspendCmdSupport & (1 << i))
			debugf("    - Unknown Bit %X:               supported", i);
	}

	debugf("Block Status Register Mask: %4.4X", pri->blkStatusRegMask);
	debugf("     - Lock Bit Active:      %s",
		pri->blkStatusRegMask & 1 ? "yes" : "no");
	debugf("     - Lock-Down Bit Active: %s",
		pri->blkStatusRegMask & 2 ? "yes" : "no");
	for (i = 2; i < 3; i++) {
		if (pri->blkStatusRegMask & (1 << i))
			debugf("    - Unknown Bit %X Active: yes", i);
	}
	debugf("    - EFA Lock Bit:         %s",
		pri->blkStatusRegMask & 16 ? "yes" : "no");
	debugf("    - EFA Lock-Down Bit:    %s",
		pri->blkStatusRegMask & 32 ? "yes" : "no");
	for (i = 6; i < 16; i++) {
		if (pri->blkStatusRegMask & (1 << i))
			debugf("    - Unknown Bit %X Active: yes", i);
	}

	debugf("Vcc Logic Supply Optimum Program/Erase Voltage: %d.%d V",
		pri->vccOptimal >> 4, pri->vccOptimal & 0xf);
	
	if (pri->vppOptimal)
		debugf("Vpp Programming Supply Optimum "
			"Program/Erase Voltage: %d.%d V",
			pri->vppOptimal >> 4, pri->vppOptimal & 0xf);

	/* free intel ext info structure, CFI info structure was freed before */
	dfree(pri);
}
#endif

/** \brief initialize flash */
static void init(struct Driver * driver) {
#ifdef DEBUG_FLASH
	flashInfo();
#endif
}

/** \brief lock/unlock/lockdown flash block
 * \param block block base address
 * \param cmd command */
static void blockLock(uint32_t block, uint16_t cmd) {
	/* clear status */
	cmdClearStatus();
	
	/* enter lock mode */
	cmdBlockLock(block);
	
	/* lock/unlock/lockdown block */
	write(block, 0x0, cmd);
	
	/* return to read array mode */
	cmdReadArray(block);
}

/** \brief lock flash block
 * \param block block base address
 * \note disable interrupts or call from <b>Interrupt context</b> when
 *  locking a block of the boot partition */
void flashBlockLock(uint32_t block) {
	blockLock(block, BLOCK_LOCK);
}

/** \brief unlock flash block
 * \param block block base address
 * \note disable interrupts or call from <b>Interrupt context</b> when
 *  unlocking a block of the boot partition */
void flashBlockUnlock(uint32_t block) {
	blockLock(block, BLOCK_UNLOCK);
}

/** \brief lockdown flash block
 * \param block block base address
 * \note disable interrupts or call from <b>Interrupt context</b> when
 *  locking down a block of the boot partition */
void flashBlockLockdown(uint32_t block) {
	blockLock(block, BLOCK_LOCKDOWN);
}

/** \brief query the lock status of a flash block
 * \param block block base address
 * \return \c TRUE iff block is locked or locked down
 * \note do not execute code from boot partition (e.g. disable interrupts) */
bool flashBlockQueryLock(uint32_t block) {
	uint8_t status;
	
	cmdReadQuery(block);
	
	status = read(block, 0x2);
	
	cmdReadArray(block);
	
	return status & (LOCK_STATUS_LOCKED | LOCK_STATUS_LOCKEDDOWN) ?
		TRUE : FALSE;
}

/** \brief erase block
 * \param block block base address
 * \return \c TRUE iff operation succeeded
 * \note do not execute code from boot partition (e.g. disable interrupts) */
bool flashBlockErase(uint32_t block) {
	uint16_t result;
	
	/* clear status */
	cmdClearStatus();
	
	/* prepare for erase */
	cmdEraseSetup(block);
	
	/* erase block */
	write(block, 0x0, CMD_ERASE_CONFIRM);
	
	/* status check and return to read array mode */
	result = readStatusWait(block);
	
	return (result & STATUS_ERASE_ERR) ? FALSE : TRUE;
}

/** write single halfword
 * \param addr destination address
 * \param data data
 * \return \c TRUE iff operation succeeded
 * \note do not execute code from boot partition (e.g. disable interrupts) */
bool flashBlockWriteShort(uint32_t addr, uint16_t data) {
	uint16_t result;
	uint32_t block = addr & ~0x1ffff;
	
	/* clear status */
	cmdClearStatus();
	
	/* prepare for programming */
	cmdProgramSetup(addr);
	
	/* write data */
	write(addr, 0x0, data);
	
	/* status check and return to read array mode */
	result = readStatusWait(block);
	
	return (result & STATUS_PGM_ERR) ? FALSE : TRUE;
}

/** write buffered data
 *
 * uses the program to buffer algorithm of the chip. It is possible to
 *  program up to 64 byte in parallel.
 *
 * \param addr destination address
 * \param data data, up to 32 half words
 * \param length number of halfwords
 * \return \c TRUE iff operation succeeded
 * \note do not execute code from boot partition (e.g. disable interrupts)
 * \note only halfwords can be programmed, thus an odd number of bytes must be
 *  combined with flashBlockWriteShort()
 * \note only write to 32 bit aligned addresses */
bool flashBlockWriteBuffered(uint32_t addr, const uint16_t * data,
	uint16_t length) {
	uint16_t result;
	uint16_t i;
	volatile uint16_t * cur;
	uint32_t block = addr & ~0x1ffff;
	
	/* clear status */
	cmdClearStatus();
	
	/* prepare for programming */
	cmdProgramBufferSetup(block);
	
	/* wait for buffer */
	do {
		result = read(block, 0x0);
	} while((result & 0x80) == 0);
	
	/* write length - 1 */
	write(block, 0x0, length - 1);
	
	/* write buffer */
	cur = ADDR16(addr, 0);
	for (i = 0; i < length; ++i)
		cur[i] = data[i];
	
	/* confirm write */
	write(block, 0x0, CMD_WRITE_TO_BUFFER_CONFIRM);
	
	/* status check and return to read array mode */
	result = readStatusWait(block);
	
	return (result & STATUS_PGM_ERR) ? FALSE : TRUE;
}

/** @} */


