/* Interrupt handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <os_arm_bare_registers_def.h>
#include <os_arm_bare_interrupt.h>
#include <inttypes.h>
#include "interrupt.h"
#include "usb/usb_interrupt.h"
#include "dma.h"
#include "pmic_interrupt.h"
#include "timer.h"

/** \file interrupt.c
 * \brief interrupt handler
 */

static void init(struct Driver * driver);
static void start(struct Driver * driver);
static void stop(struct Driver * driver);
static void finish(struct Driver * driver);

/** \brief lock counter as defined in os_arm_bare_memory.c */
extern int8_t lock_counter;

/** \brief interrupt driver structure */
struct Driver interruptDriver = {
	.name = "Interrupt",
	
	.init = &init,
	.start = &start,
	.stop = &stop,
	.finish = &finish
};

/** \brief initialize interrupts
 *
 * sets the interrupt priorities */
static void init(struct Driver * driver) {
	uint8_t i;
	uint32_t mask1, mask2;
	
	/* disable interrupts */
	SENF_DISABLE_INTERRUPTS();
	/* store interrupt registers */
	mask1 = ICMR;
	mask2 = ICMR2;
	/* reset interrupt registers */
	ICMR = 0;
	ICMR2 = 0;
	/* set priorities */
	for (i = 0; i <= 39; ++i)
		IPR(i) = i | IPR_VAL;
	
	/* restore interrupt registers */
	ICMR = mask1;
	ICMR2 = mask2;
	
	/* generate pmic interrupts */
	/* TODO: let gpio driver do this! */
	ICMR_REG(IID_GPIO_1) |= INTERRUPT_BIT(IID_GPIO_1);
	GPDR(1) &= ~GPIO_BIT(1);
	GFER(1) |= GPIO_BIT(1);
}

/** \brief enable interrupts */
static void start(struct Driver * driver) {
	/* enable interrupts */
	SENF_ENABLE_INTERRUPTS();
}

/** \brief disable interrupts */
static void stop(struct Driver * driver) {
	/* disable interrupts */
	SENF_DISABLE_INTERRUPTS();
}

/** \brief mask interrupts */
static void finish(struct Driver * driver) {
	int i;
	
	/* mask all interrupts */
	ICMR = 0;
	ICMR2 = 0;
	
	/* mask all interrupts */
	for (i = 0; i <= 39; ++i) {
		IPR(i) &= ~IPR_VAL;
	}
}

/** \brief interrupt handler
 * \note the interrupt handler disables and enables the IRQ interrupt
 *  automatically */
__attribute__((interrupt("IRQ")))
void hplarmv_irq() {
	uint32_t IRQPending;
	
	lock_counter++;
	
	IRQPending = ICHP; // Determine which interrupt to service
	IRQPending >>= 16; // Right justify to the IRQ portion

	while (IRQPending & (1 << 15)) {
		uint8_t PeripheralID = (IRQPending & 0x3f); // Get rid of the Valid bit
		switch (PeripheralID) {
			case IID_GPIO_1:
				pmicInterrupt();
				break;
			case IID_USBC:
				usbInterrupt();
				break;
			case IID_DMAC:
				dmaInterrupt();
				break;
			case IID_OST_4_11:
				timerInterrupt();
				break;
			default:
				break;
		}
		IRQPending = ICHP;  // Determine which interrupt to service
		IRQPending >>= 16;  // Right justify to the IRQ portion
	}
	
	if (--lock_counter < 0)
		lock_counter = 0;
}

