/* CPU Frequency Scaling
 * Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern
 *
 * This file features some ideas and code of the pxa2xx cpufreq driver of the
 * Linux kernel.
 *
 *  Copyright (C) 2002,2003 Intrinsyc Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "inttypes.h"
#include "os_arm_bare_interrupt.h"
#include "os_arm_bare_cpufreq.h"
#include "os_arm_bare_registers_def.h"
#include "pmic.h"

/** current core frequency */
static enum SEnF_ARM_CPUFREQ currentFrequency;
/** previous core frequency (used by deep idle) */
enum SEnF_ARM_CPUFREQ oldCoreFrequency;

/** sdram rows, needed for DRI calculation */
static uint32_t sdramRows;

/** initialize sdramRows variable */
static inline void sdramRowsInit() {
	uint32_t mdcnfg = MDCNFG;
	uint32_t drac2 = 0, drac0 = 0;

	if (mdcnfg & (MDCNFG_DE2 | MDCNFG_DE3))
		drac2 = (mdcnfg >> 21) & 0x3;

	if (mdcnfg & (MDCNFG_DE0 | MDCNFG_DE1))
		drac0 = (mdcnfg >> 5) & 0x3;

	sdramRows = 1 << (11 + (drac0 > drac2 ? drac0 : drac2));
}

/** calculate sdram refresh value for a given memory frequency */
static inline uint32_t mdrefrGetDri(uint32_t freq) {
	return (freq * 64 / (sdramRows - 31)) / 32;
}

/** finally set the cpu frequency.
 * this needs to be executed from sram
 * \param preset_mdrefr set this value into sdram configuration register BEFORE
 *  the frequency change
 * \param postset_mdrefr set this value into sdram configuration register AFTER
 *  the frequency change
 * \param cccr new CCCR value
 * \param cclkcfg value to be set into CCLKCFG co-processor register */
#ifdef XENV
__attribute__((section(".ramtext"),long_call))
#else
__attribute__((section(".ramtext"),long_call,no_instrument_function))
#endif
static void changeFrequency(uint32_t preset_mdrefr, uint32_t postset_mdrefr,
	uint32_t cccr, uint32_t cclkcfg) {
	/* XXX: do we need to set the flash in async mode? */

	/* configure cccr register */
	CCCR = cccr;
	asm volatile(
		/* read mdrefr register (ensures some stabilization) */
		"ldr r4, [%[mdrefr]]\n\t"
		/* now do some tricky alignment jumps */
		"b 2f\n\t"
		".align 5\n\t"
		"1:\n\t"
		/* set the preset mdrefr value */
		"str %[preset_mdrefr], [%[mdrefr]]\n\t"
		/* do the frequency change */
		"mcr p14, 0, %[cclkcfg], c6, c0, 0\n\t"
		/* set the postset mdrefr value */
		"str %[postset_mdrefr], [%[mdrefr]]\n\t"
		"b 3f\n"
		"2: b 1b\n\t"
		"3: nop"
		:
		: [mdrefr] "r" (&MDREFR), [cclkcfg] "r" (cclkcfg),
		  [preset_mdrefr] "r" (preset_mdrefr),
		  [postset_mdrefr] "r" (postset_mdrefr)
		: "memory", "r4");

	/* XXX: do we need to set the flash back into sync mode? */
}

#define MDREFR_DB2_MASK         (MDREFR_K2DB2 | MDREFR_K1DB2)

/** Set the frequency of the CPU
 * \param freq frequency, one of the enums values
 */
void cpufreqSetFrequency(enum SEnF_ARM_CPUFREQ freq) {
	uint32_t new_freq_cpu, new_freq_mem;
	uint32_t new_vcc_core;
	uint32_t cccr, cclkcfg;
	uint32_t preset_mdrefr, postset_mdrefr;
	bool div2;

	if (currentFrequency == freq) {
		/* do nothing if there is nothing to do */
		return;
	}

	/* lookup new register values for the given frequency */
	switch (freq) {
		case SEnF_ARM_CPUFREQ_13:
			cccr = CCCR_CPDIS | CCCR_A;
			cclkcfg = CLKCFG_F;
			new_freq_cpu = 13000;
			new_freq_mem = 13000;
			new_vcc_core = 950000; /* 950 mV */
			div2 = FALSE;
			break;

		case SEnF_ARM_CPUFREQ_104:
			cccr = CCCR_L (8) | CCCR_2N (2) | CCCR_A;
			cclkcfg = CLKCFG_F | CLKCFG_B | CLKCFG_T;
			new_freq_cpu = 104000;
			new_freq_mem = 104000;
			new_vcc_core = 950000; /* 950 mV */
			div2 = FALSE;
			break;

		case SEnF_ARM_CPUFREQ_208:
			cccr = CCCR_L (16) | CCCR_2N (2) | CCCR_A;
			cclkcfg = CLKCFG_F | CLKCFG_B | CLKCFG_T;
			new_freq_cpu = 208000;
			new_freq_mem = 208000;
			new_vcc_core = 1200000; /* 1.2 V */
			div2 = TRUE;
			break;

		case SEnF_ARM_CPUFREQ_416:
			cccr = CCCR_L (16) | CCCR_2N (4) | CCCR_A;
			cclkcfg = CLKCFG_F | CLKCFG_B | CLKCFG_T;
			new_freq_cpu = 416000;
			new_freq_mem = 208000; /* this is not a typo */
			new_vcc_core = 1350000; /* 1.35 V */
			div2 = TRUE;
			break;

		default:
			/* frequency not supported */
			return;
	}

	/*debugf("Setting frequency to %d, mem to %d and vcc to %d",
		new_freq_cpu, new_freq_mem, new_vcc_core);*/

	/* setup voltage change */
	uint32_t vcc_regval;

	switch (new_vcc_core) {
		case 950000:
			vcc_regval = B2R1_TRIM_P95_V;
			break;

		case 1200000:
			vcc_regval = B2R1_TRIM_1P2_V;
			break;

		case 1350000:
			vcc_regval = B2R1_TRIM_1P35_V;
			break;

		default:
			/* fallback: set to highest (used) voltage */
			vcc_regval = B2R1_TRIM_1P35_V;
			break;
	}
	//XXX
	//vcc_regval = B2R1_TRIM_1P35_V;

	PVCR = PVCR_Read_Pointer (0) | PVCR_Slave_Address (PMIC_SLAVE_ADDR);
	//write I2C commands to PCMDx Register
	// MBC = MultiByteCommand LC = LastCommand DCE = DelayCommandExecution
	PCMD0 = PCMDx_SQC (0) | PCMDx_DCE | PCMDx_MBC  | PCMDx_LC | PCMDx_Command_Data (PMIC_BUCK2_REG1);
	PCMD1 = PCMDx_SQC (0) | PCMDx_DCE | PCMDx_LC |
		PCMDx_Command_Data ((B2R1_TRIM_MASK & vcc_regval)| B2R1_GO);

	/* calculate new values for mdrefr (memory refresh configuration) */

	preset_mdrefr = postset_mdrefr = MDREFR;
	postset_mdrefr = (postset_mdrefr & ~MDREFR_DRI_MASK) |
		mdrefrGetDri(new_freq_mem);
	if ((preset_mdrefr & MDREFR_DRI_MASK) > mdrefrGetDri(new_freq_mem)) {
		preset_mdrefr = postset_mdrefr;
	}

	if (div2) {
		preset_mdrefr |= MDREFR_K1DB2 | MDREFR_K2DB2;
		postset_mdrefr |= MDREFR_K1DB2 | MDREFR_K2DB2;
	} else {
		postset_mdrefr &= ~(MDREFR_K1DB2 | MDREFR_K2DB2);
	}

	//debugf("params: %08x %08x %08x %d", MDREFR, preset_mdrefr, postset_mdrefr, sdramRows);

	/* and now...the actual thing */
	SENF_DISABLE_INTERRUPTS();
	changeFrequency(preset_mdrefr, postset_mdrefr, cccr, cclkcfg);
	currentFrequency = freq;
	SENF_ENABLE_INTERRUPTS();
}

/** returns current frequency of the CPU
 * \return current cpu frequency
 */
enum SEnF_ARM_CPUFREQ cpufreqGetFrequency() {
	return currentFrequency;
}

void cpufreqInit() {
	sdramRowsInit();
	currentFrequency = SEnF_ARM_CPUFREQ_13;
	#ifdef IMOTE_MAX_SPEED
		cpufreqSetFrequency(SEnF_ARM_CPUFREQ_416);
	#else
		cpufreqSetFrequency(SEnF_ARM_CPUFREQ_104);
	#endif
}
