/* Simple Bottom Half Task Scheduler
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <os_arm_bare_interrupt.h>
#include "dmalloc.h"
#include "task.h"

/** \file task.c
 * \brief adding and executing bottom half routines
 */

/** \brief current task */
static struct Task * currentTask = NULL;

/** \brief run all pending tasks and unlink them from the list
 *
 * when this function returns, no further tasks are enqueued
 * \note do not call from <b>Interrupt context</b> (for obvious reasons)
 * \note \a no_instrument_function as this function is running too often */
__attribute__((no_instrument_function))
void taskRun() {
	struct Task * task;
	
	/* iterate over tasks */
	for (task = currentTask; task != NULL; task = currentTask) {
		if (task->interruptible) {
			/* task is interruptible -> run it and disable interrupts then */
			task->task();
			SENF_DISABLE_INTERRUPTS();
		} else {
			/* task is uninterruptible -> disable interrupts and run task then */
			SENF_DISABLE_INTERRUPTS();
			task->task();
		}
		
		/* interrupts are disabled: unlink task */
		currentTask = task->next;
		free(task);
		/* reenable interrupts */
		SENF_ENABLE_INTERRUPTS();
	}
}

/** \brief add a new task
 * \param task the bottom half routine
 * \param interruptible set to FALSE iff this function may not be interrupted
 * \note of course, this routine can safely be called from
 *  <b>Interrupt context</b> */
void taskAdd(void (*task)(), bool interruptible) {
	struct Task * newTask, * lastTask;
	
	/* allocate new task */
	/* as malloc() will disable interrupts (and maybe reenable them afterwards)
		and we need interrupts to be disabled afterwards, disable interrupts
		now to save an enable/disable cycle */
	SENF_DISABLE_INTERRUPTS();
	newTask = (struct Task*)dmalloc(sizeof *newTask);
	newTask->task = task;
	newTask->interruptible = interruptible;
	newTask->next = NULL;
	
	/* link task */
	if (currentTask == NULL) {
		currentTask = newTask;
	} else {
		/* get the last task */
		for (lastTask = currentTask; lastTask->next != NULL;
			lastTask = lastTask->next);
		/* append */
		lastTask->next = newTask;
	}
	/* reenable interrupts */
	SENF_ENABLE_INTERRUPTS();
}

/** \brief check if a task is pending
 * \return \c TRUE iff a task is pending
 * \note this routine should be run in <b>Interrupt context</b> or
 *  when interrupts are \b disabled. If not, two consecutive calls
 *  could yield different results, thus the value is only reasonable when
 *  the constraints above apply */
bool taskPending() {
	return currentTask != NULL;
}

