/* Timer Driver
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "timer.h"

/** \defgroup timer Timer Subsystem
 * @{
 */

/** \file timer.c
 * \brief OS Timer 4-11 routines
 */

/** callback type */
typedef void(*timerCallback)(uint8_t timerId);
/** timer callbacks */
static timerCallback callbacks[11];

static void start(struct Driver * driver);
static void stop(struct Driver * driver);
static void finish(struct Driver * driver);

/** \brief timer driver structure */
struct Driver timerDriver = {
	.name = "Driver",
	
	/* init is the same as finish */
	.init = &finish,
	.start = &start,
	.stop = &stop,
	.finish = &finish
};

/** \brief start timer driver */
static void start(struct Driver * driver) {
	/* enables the 4-11 timer interrupt */
	ICMR_REG(IID_OST_4_11) |= INTERRUPT_BIT(IID_OST_4_11);
}

/** \brief stop timer driver */
static void stop(struct Driver * driver) {
	/* disable timer interrupt */
	ICMR_REG(IID_OST_4_11) &= ~INTERRUPT_BIT(IID_OST_4_11);
}

/** \brief clean up timers */
static void finish(struct Driver * driver) {
	timerStopAll();
}

/** \brief stop all timers
 *
 * stops all timers and clears the interrupt */
void timerStopAll() {
	uint8_t timerId;
	
	/* stop all timers */
	for (timerId = 4; timerId <= 11; ++timerId)
		timerStop(timerId);
	
	/* clear pending interrupts */
	OSSR |= 0;
}

/** \brief setup a timer (counter)
 * \param timerId timer id (4 - 11)
 * \param resolution timer resolution
 *  - 1: 1/32768th of a second
 *  - 2: 1 ms
 *  - 3: 1 s
 *  - 4: 1 us
 * \param compareBase compare the counter of the base timer against this
 *  timers matching register. The base of channel 4-7 is channel 4 and the base
 *  of channel 8-11 is channel 8.
 * \param periodic \c TRUE iff channel should continue incrementing after a
 *  match. Matches are used in waiting timers only, see timerWait()
 * \param reset \c TRUE iff channel should be reset after a match
 * \see timerWait
 * \see timerGet */
 void timerSet(uint8_t timerId, uint8_t resolution, bool compareBase,
	bool periodic, bool reset) {
	timerStop(timerId);
	OMCR4_11(timerId) = (compareBase ? 0 : OMCR_C) | (periodic ? OMCR_P : 0) |
		(reset ? OMCR_R : 0) | OMCR_CRES(resolution);
}

/** \brief start a timer
 * \param timerId timer id (4 - 11)
 * \param start starting value
 * \see timerSet
 */
void timerStart(uint8_t timerId, uint32_t start) {
	OSCR4_11(timerId) = start;
}

/** \brief stop a timer
 * \param timerId timer id (4 - 11)
 * \see timerStart
 * \see timerWait
 */
void timerStop(uint8_t timerId) {
	/* stop interrupt */
	OIER &= ~OIER_E(timerId);
	/* stop timer */
	OMCR4_11(timerId) &= (~(OMCR_C | OMCR_CRES(3)) & 0xff);
	OMCR4_11(timerId) |= OMCR_C;
}

/** \brief setup a timer with callback
 * \param timerId timer id (4 - 11)
 * \param resolution timer resolution
 *  - 1: 1/32768th of a second
 *  - 2: 1 ms
 *  - 3: 1 s
 *  - 4: 1 us
 * \param compareBase compare the counter of the base timer against this
 *  timers matching register. The base of channel 4-7 is channel 4 and the base
 *  of channel 8-11 is channel 8.
 * \param periodic \c TRUE iff channel should continue incrementing after a
 *  match
 * \param reset \c TRUE iff channel should be reset after a match
 * \param duration duration after which the timeout occures
 * \param cbTimer callback on timer match */
void timerWait(uint8_t timerId, uint8_t resolution, bool compareBase,
	bool periodic, bool reset,
	uint32_t duration, void(*cbTimer)(uint8_t timerId)) {
	/* setup timer */
	timerSet(timerId, resolution, compareBase, periodic, reset);
	/* setup match register */
	OSMR4_11(timerId) = duration;
	/* setup callback */
	callbacks[timerId] = cbTimer;
	/* setup interrupt */
	OIER |= OIER_E(timerId);
	/* start timer */
	timerStart(timerId, 0);
}

/** \brief timer interrupt
 * called by interrupt handler when a timer interrupt occured
 * \note called in <b>Interrupt Context</b> */
void timerInterrupt() {
	uint32_t ossr = OSSR;
	uint8_t timerId = 4;
	
	OSSR = ossr & ~0xf;
	ossr >>= timerId;
	while (ossr) {
		if (ossr & 1)
			(*callbacks[timerId])(timerId);
		
		ossr >>= 1;
		timerId++;
	}
}

/** @} */

