/* USB Driver: Dynamic Configuration
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef USB_CONFIGURATION_H
#define USB_CONFIGURATION_H

#include <inttypes.h>
#include "usb_config.h"
#include "../dma.h"

/** \addtogroup usbconfig
 * @{
 */

/** \file usb_configuration.h
 * \brief USB Configuration
 * \note see usb_config.h for static configuration
 */

struct UsbRequest;

/** \brief USB Endpoint Type
 * \details this is also known as the transfer type */
enum UsbEndpointType {
	USB_ENDPOINT_TYPE_CONTROL,
	USB_ENDPOINT_TYPE_ISOCHRONOUS,
	USB_ENDPOINT_TYPE_BULK,
	USB_ENDPOINT_TYPE_INTERRUPT
};

/** \brief USB Endpoint Direction */
enum UsbEndpointDirection {
	/** \brief from Host to Client */
	USB_ENDPOINT_DIR_IN,
	/** \brief from Client to Host */
	USB_ENDPOINT_DIR_OUT
};

/** \brief USB Endpoint definition */
struct UsbEndpoint {
	/** \brief endpoint address */
	uint8_t epAddress;
	/** \brief endpoint transfer type */
	enum UsbEndpointType epType;
	/** \brief endpoint direction */
	enum UsbEndpointDirection epDirection;
	/** \brief endpoint max packet size */
	uint16_t epMaxPacketSize;
	/** \brief true iff double buffering should be used for that endpoint */
	bool doubleBuffer;
	/** \brief true iff dma should be used for that endpoint */
	bool useDma;
	
	/** \brief udc endpoint number */
	uint8_t udcEndpoint;
	
	/** \brief dma channel */
	uint8_t dmaChannel;
	/** \brief dma data descriptor */
	struct DmaDescriptor * dmaDataDesc;
	/** \brief dma end descriptor */
	struct DmaDescriptor * dmaEndDesc;
	/** \brief dma running (request pending) */
	bool dmaRunning;
	
	/** \brief true iff endpoint is enabled */
	bool enabled;
	/** \brief true iff endpoint is handled at the moment.
	 *  used to avoid recurions */
	bool inHandle;
	
	/** \brief rx/tx requests */
	struct UsbRequest * req;
};

/** \brief USB Alternate Setting definition */
struct UsbAltSetting {
	/** \brief Class of alternate setting */
	uint8_t altClass;
	/** \brief SubClass of alternate setting */
	uint8_t altSubClass;
	/** \brief Protocol of alternate setting */
	uint8_t altProtocol;
	/** \brief optional description of alternate setting
	 * \note can be \c NULL */
	char * altDescription;
	
	/** \brief number of endpoints in this setting */
	uint8_t numEndpoints;
	/** \brief endpoints */
	struct UsbEndpoint * endpoints[USB_CONFIG_MAX_ENDPOINTS_PER_ALTSETTING];
};

/** \brief USB Interface definition */
struct UsbInterface {
	/** \brief number of alternate settings */
	uint8_t numAltSettings;
	/** \brief alternate settings */
	struct UsbAltSetting * altSettings[USB_CONFIG_MAX_ALTSETTINGS_PER_INTERFACE];
};

/** \brief USB Configuration definition
 * \note do not intermix with ::UsbDevice Configuration */
struct UsbConfiguration {
	/** \brief \c TRUE iff powered by bus */
	bool confSelfPowered;
	/** \brief \c TRUE iff device supports remote wakeups
	 * \note not further implemented */
	bool confRemoteWakeup;
	/** \brief power consumption in \c mA */
	uint16_t confMaxPower;
	/** \brief optional configuration description
	 * \note can be \c NULL */
	char * confDescription;
	
	/** \brief number of interfaces */
	uint8_t numInterfaces;
	/** \brief interfaces */
	struct UsbInterface * interfaces[USB_CONFIG_MAX_INTERFACES_PER_CONFIGURATION];
};

/** \brief USB Device Configuration */
struct UsbDevice {
	/** \brief Device Class
	 * \note \c 0 means defined at interface (resp. alt setting) level */
	uint8_t devClass;
	/** \brief Device SubClass
     * \note \c 0 means defined at interface (resp. alt setting) level if
     *  devClass is also \c 0 */
	uint8_t devSubClass;
	/** \brief Device Protocol
	 * \note \c 0 means defined at interface (resp. alt setting) level */
	uint8_t devProtocol;
	/** \brief Device vendor id */
	uint16_t devIdVendor;
	/** \brief Device product id */
	uint16_t devIdProduct;
	/** \brief Device version in \c BCD */ 
	uint16_t devBCDDevice;
	/** \brief optional Device Manufacturer String
	 * \note can be \c NULL */
	char * devManufacturer;
	/** \brief optional Device Product String
	 * \note can be \c NULL */
	char * devProduct;
	/** \brief optional Device Serial Number String
	 * \note can be \c NULL */
	char * devSerialNumber;
	
	/** \brief String Language ID
	 * \note Only one language is supportet at the moment */
	uint16_t languageID;
	
	/** \brief callback for bus reset
	 * \details called whenever bus is reset by host
	 * \note can be \c NULL if not used */
	void (*cbBusReset)(const struct UsbDevice *devConf);
	/** \brief callback for bus configuration
	 * \details called whenever bus is (re)configured by host
	 * \note can be \c NULL if not used */
	void (*cbConfigured)(const struct UsbDevice *devConf,
		uint8_t configIdx, uint8_t ifaceIdx, uint8_t settingIdx);
	
	/** \brief number of configurations */
	uint8_t numConfigurations;
	/** \brief configurations */
	struct UsbConfiguration * configurations[];
};

void usbConfigLoad(struct UsbDevice * devConf);

/** @} */

#endif

