/* USB Driver: Endpoint handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef USB_ENDPOINT_H
#define USB_ENDPOINT_H

#include <inttypes.h>
#include <os_arm_bare_registers_def.h>
#include "usb_configuration.h"
#include "usb_request.h"
#include "usb_control_pipe_internal.h"
#include "../assert.h"

/** \cond INTERNAL */

/** \addtogroup usbcomm
 * @{
 */
 
/** \file usb_endpoint_internal.h
 * \brief USB Endpoint handling
 */

void usbEndpointInterrupt();

void usbEndpointNuke(struct UsbEndpoint * ep, enum UsbRequestStatus status);

/** \brief tests wether an OUT endpoint is empty
 * \param ep endpoint
 * \return true iff endpoint is empty */
static inline bool usbEndpointIsEmpty(const struct UsbEndpoint * ep) {
	ASSERT(usbIsControlPipe(ep) || ep->epDirection == USB_ENDPOINT_DIR_OUT);
	
	if (usbIsControlPipe(ep))
		return !(UDCCSR0 & UDCCSR0_RNE);
	else
		return !(UDCCSRAX(ep->udcEndpoint) & UDCCSRAX_BNE);
}

/** \brief tests wether an IN endpoint is full
 * \param ep endpoint
 * \return true iff endpoint is full */
static inline bool usbEndpointIsFull(const struct UsbEndpoint * ep) {
	ASSERT(usbIsControlPipe(ep) || ep->epDirection == USB_ENDPOINT_DIR_IN);
	
	if (usbIsControlPipe(ep))
		return !!(UDCCSR0 & UDCCSR0_IPR);
	else
		return !(UDCCSRAX(ep->udcEndpoint) & UDCCSRAX_BNF);
}

/** \brief tests wether an OUT endpoint has packets
 * \param ep endpoint
 * \return true iff endpoint has packets */
static inline bool usbEndpointHasPkt(const struct UsbEndpoint * ep) {
	ASSERT(usbIsControlPipe(ep) || ep->epDirection == USB_ENDPOINT_DIR_OUT);
	
	if (usbIsControlPipe(ep))
		return !!(UDCCSR0 & UDCCSR0_OPC);
	else
		return !!(UDCCSRAX(ep->udcEndpoint) & UDCCSRAX_PC);
}

/** \brief count remaining bytes of an OUT endpoint
 * \param ep endpoint
 * \return available bytes in the fifo buffer */
static inline uint32_t usbEndpointBytesRemain(const struct UsbEndpoint * ep) {
	return UDCBCR(ep->udcEndpoint) & 0x3ff;
}

uint32_t usbEndpointWritePacket(struct UsbEndpoint * ep,
	struct UsbRequest * req);

uint32_t usbEndpointReadPacket(struct UsbEndpoint * ep,
	struct UsbRequest * req);

void usbEndpointsConfigure(struct UsbAltSetting * newAltsetting);

void usbEndpointsFinishAll(enum UsbRequestStatus status);

void usbEndpointHandleData(struct UsbEndpoint * ep);

/** @} */

/** \endcond */

#endif

