/* USB Driver: Main Interface
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef USB_MAIN_INTERNAL_H
#define USB_MAIN_INTERNAL_H

#include <stdlib.h>
#include "usb_config.h"
#include "usb_configuration.h"
#include "usb_control_pipe_internal.h"

/** \cond INTERNAL */

/** \addtogroup usb
 * @{
 */
 
/** \file usb_main_internal.h
 * \brief USB Internals
 */

/** \brief global udc status structure */
struct UsbUdc {
	/** \brief UDC Device configuration */
	struct UsbDevice * config;
	/** \brief current alternate interface setting (holds current endpoint
	 *  description)
	 * \note can be NULL if bus is unconfigured */
	struct UsbAltSetting * curAltSetting;

	/** \brief current config index */
	uint8_t configIdx;
	/** \brief current interface index */
	uint8_t ifaceIdx;
	/** \brief current alt setting index */
	uint8_t settingIdx;
	
	/** \brief control pipe state */
	enum USB_CONTROL_PIPE_STATE cpState;
	
	/** \brief endpoint mapping: maps an endpoint index of the udc into
	 *  the current configuration */
	struct UsbEndpoint * ep[USB_CONFIG_NUM_ENDPOINTS];
	
	/** \brief number of used udc endpoints, determined by the device
	 *  configuration */
	uint32_t endpointsUsed;
	
	/** \brief number of currently used strings, determined by the device
	 *  configuration */
	uint8_t stringsUsed;
	
	/** \brief descriptor string map: maps a string index into the
	 *  current configuration */
	const char * descriptorStrings[USB_CONFIG_MAX_STRINGS];
};

extern struct UsbUdc usbUdc;

/** \brief write control/status register of any endpoint
 * \param ep endpoint
 * \param val value */
static inline void usbWrite_UDCCSR(struct UsbEndpoint * ep, uint32_t val) {
	/* no special treatment of ep0, because it fits the following code */
	UDCCSRAX(ep->udcEndpoint) = val;
}

/** \brief set bits in control register of udc
 * \param mask mask to be set in control register */
static inline void usbSet_UDCCR(uint32_t mask) {
	uint32_t udccr = UDCCR;
	UDCCR = (udccr & UDCCR_MASK_BITS) | (mask & UDCCR_MASK_BITS);
}

/** \brief clear bits in control register of udc
 * \param mask mask to be cleared in control register */
static inline void usbClear_UDCCR(uint32_t mask) {
	uint32_t udccr = UDCCR;
	UDCCR = (udccr & UDCCR_MASK_BITS) & ~(mask & UDCCR_MASK_BITS);
}


/** @} */

/** \endcond */

#endif

