/* USB Driver: Interface for request handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include "usb_request.h"
#include "usb_configuration.h"
#include "usb_endpoint_internal.h"
#include "usb_request_internal.h"
#include "usb_control_pipe_internal.h"
#include "usb_main_internal.h"
#include "usb_debug.h"
#include <os_arm_bare_interrupt.h>
#include "../assert.h"
#include "../core/cache.h"

/** \addtogroup usbcomm
 * @{
 */

/** \file usb_request.c
 * \brief USB Requests
 *
 * setup, enqueue and dequeue requests on endpoints
 */

/** \brief setup a communication request
 * \param [out] req request to be setup
 * \param ep endpoint to be used for request
 * \param data data buffer
 * \param length data buffer length
 * \param callback completion function, can be NULL */
void usbRequestFill(struct UsbRequest * req, struct UsbEndpoint * ep,
	void * data, uint32_t length, UsbRequestCallback callback) {
	req->ep = ep;
	req->data = data;
	req->length = length;
	req->callback = callback;
	req->inUse = FALSE;
}

/** \brief enqueue a communication request on a specific endpoint
 * \param req request to be enqueued, must be idle (i.e. not linked)
 * \return true iff request could be enqueued */
bool usbRequestEnqueue(struct UsbRequest * req) {
	bool recursionDetected, first;
	struct UsbEndpoint * ep = req->ep;
	
	ASSERT(ep->enabled);
	ASSERT(!req->inUse);
	
	if (ep->useDma) {
		ASSERT(!((uint32_t)req->data & 0x1f));
	}
	
	SENF_DISABLE_INTERRUPTS();
	
	req->actual = 0;
	req->inUse = TRUE;
	req->next = NULL;
	
	recursionDetected = ep->inHandle;
	
	first = ep->req == NULL;
	
	usbEPInfo(ep, "Queueing request %p(first=%s), len %d, buf %p",
		req, first ? "yes" : "no", req->length, req->data);
	
	/* enqueue request in request list */
	struct UsbRequest ** last;
	/* get pointer to the last requests next field */
	for (last = &ep->req; *last; last = &((*last)->next));
	/* assign the new request to that field */
	*last = req;
	
	if (!ep->useDma || ep->epDirection == USB_ENDPOINT_DIR_OUT)
		UDCICR(ep->udcEndpoint) |= UDC_INT(ep->udcEndpoint, UDCICR_IE);
	
	if (usbIsControlPipe(ep)) {
		switch(usbUdc.cpState) {
		case USB_CP_IN_DATA_STAGE:
			if (!usbEndpointIsFull(ep))
				if (usbControlPipeWrite(req))
					usbControlInRequestComplete(req);
			break;
		case USB_CP_OUT_DATA_STAGE:
			if (req->length == 0 || !usbEndpointHasPkt(ep))
				if (usbControlPipeRead(req))
					usbControlOutRequestComplete(req);
			break;
		default:
			usbCPErr("Odd state to send request on EP0: %d", usbUdc.cpState);
			usbRequestComplete(ep, req, USB_REQUEST_STATUS_CANCELED);
			break;
		}
	} else {
		if (ep->useDma)
			usbHandleDmaRequest(ep);
		else if (!recursionDetected)
			usbEndpointHandleData(ep);
	}
	
	SENF_ENABLE_INTERRUPTS();
	
	return TRUE;
}

void usbHandleDmaRequest(struct UsbEndpoint * ep) {
	struct UsbRequest * req;
	
	req = ep->req;
	if (!req || ep->dmaRunning) return;
	
	ep->dmaRunning = TRUE;
	
	usbEPDebug(ep, "DMA Endpoint: setting up DMA request of length %d",
		req->length);
	
	if (ep->epDirection == USB_ENDPOINT_DIR_OUT) {
		/* receiving endpoint */
		dmaSetupHWToMem(ep->dmaChannel, (void*)&UDCDR(ep->udcEndpoint),
			req->data, 3, 3, req->length);
	} else {
		if (req->length != NULL)
			dCacheCleanAndInvalidateLine(req->data, req->length);
		/* sending endpoint */
		if (req->length == 0) {
			/* zlp */
			dmaSetupDescStart(ep->dmaChannel, ep->dmaEndDesc);
		} else if (req->length % ep->epMaxPacketSize) {
			/* last packet is a short packet */
			dmaSetupDescMemToHW(ep->dmaDataDesc, FALSE, req->data,
				(void*)&UDCDR(ep->udcEndpoint), 3, 3, req->length);
			dmaSetupDescSetNextDesc(ep->dmaDataDesc, ep->dmaEndDesc);
			dmaSetupDescStart(ep->dmaChannel, ep->dmaDataDesc);
		} else {
			dmaSetupHWToMem(ep->dmaChannel, req->data,
				(void*)&UDCDR(ep->udcEndpoint), 3, 3, req->length);
		}
	}
}

/** \brief cancel a request manually
 * \param req request to be cancelled */
void usbRequestCancel(struct UsbRequest * req) {
	SENF_DISABLE_INTERRUPTS();
	usbRequestComplete(req->ep, req, USB_REQUEST_STATUS_CANCELED);
	SENF_ENABLE_INTERRUPTS();
}

/** \cond INTERNAL */

/** \brief delete a request from an endpoint
 * \param ep endpoint
 * \param req request to be deleted */
static inline void requestDelete(struct UsbEndpoint * ep,
	struct UsbRequest * req) {
	struct UsbRequest ** prev;
	/* search in linked list for the pointer to the request */
	for (prev = &ep->req; *prev && *prev != req; prev = &((*prev)->next));
	/* and unlink it */
	if (*prev)
		*prev = req->next;
	
	/* if there is no more request, clear the interrupt sources for the ep */
	if (!ep->req && !usbIsControlPipe(ep)) {
		UDCICR(ep->udcEndpoint) &= ~UDC_INT(ep->udcEndpoint, UDCICR_IE);
		
		if (ep->useDma) {
			dmaKillChannel(ep->dmaChannel);
			dmaRemapChannel(ep->dmaChannel);
		}
	}
}

/** \brief complete a request on an endpoint
 * \param ep endpoint
 * \param req request to be completed
 * \param status status with which the request is completed */
void usbRequestComplete(struct UsbEndpoint * ep, struct UsbRequest * req,
	enum UsbRequestStatus status) {
	usbEPInfo(ep, "completing Request %p with status %d", req, status);
	
	/* dequeue request */
	requestDelete(ep, req);
	req->inUse = FALSE;
	req->status = status;
	if (req->callback)
		req->callback(req);
}

/** \endcond */

/** @} */

