/* Common Utils
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <os_arm_bare_interrupt.h>
#include "utils.h"
#include "core/cache.h"
#include "core/mmu.h"
#include "interrupt.h"
#include "usb/usb_main.h"
#include "dma.h"
#include "debug.h"
#include "led.h"
#include "config.h"
#include "timer.h"
#include "flash.h"
#include "sdram.h"
#include "driver.h"

/** \file utils.c
 * \brief utils for recognizing usb power, booting, rebooting, ...
 */

static void flashLed(uint8_t timerId);

/** \brief do the actual boot
 *
 * \note \a noreturn as routine does never return
 * \note this routine is executed directly from flash memory
 *
 * \param addr target address to jump to */
__attribute__((noreturn,section(".start"),no_instrument_function))
static void doBoot(register const void * addr) {
	//sdramDisable();
	
	/* load the address into a register and move the register content into
	   the program counter */
	(*(void(*)())addr)();
	/* if the image is really broken (and returns here), loop endlessly */
	error();
}

void finishDrivers() {
	driverFinish(&usbDriver);
	
	driverFinish(&dmaDriver);
	
	driverFinish(&interruptDriver);
	
	driverFinish(&timerDriver);
}

void bootPrepare() {
	uint32_t block;
	
	finishDrivers();
	
	/* lock down first 5 blocks
		TODO: if more blocks are locked, the increment must be rechecked
		because the flash block size changes then */
	for (block = 0; block <= 0x20000; block += 0x8000) {
		flashBlockLockdown(block);
	}
	
	/* clear all leds */
	ledOff(LED_RED);
	ledOff(LED_BLUE);
	ledOff(LED_GREEN);
	
	/* disable instruction cache */
	iCacheDisable();
	/* disable mmu. this will also clean and disable the data cache */
	mmuDisable();
}

/** \brief revert all settings from the boot loader and boot the code.
 *
 * The boot procedure is split in two halves. The first is exectued in sdram,
 *  cleans up all drivers and locks down the flash. The second half is executed
 *  from flash. It disables the sdram and jumps into the images entry
 * \note \a noreturn as routine does never return */
__attribute__((noreturn))
void bootTo(void (*entry)()) {
	bootPrepare();
	
	debugf("Boot...");
	
	/* do the jump */
	register volatile __typeof__(&doBoot) r = &doBoot;
	(*r)(entry);
}

/** \brief reboot imote
 * \note routine does never return */
__attribute__((noreturn,naked))
void reboot() {
	driverFinish(&usbDriver);
	
	/* enable WatchDog */
	OWER = 1;
	OSMR3 = OSCR0 + 0x1000;
	while (1);
}

/** \brief start boot timer
 * \param cb callback after timeout
 * \note uses timer 5
 * \see BOOT_TIMEOUT
 */
void utilStartBootTimeout(void(*cb)(uint8_t)) {
	if (BOOT_TIMEOUT != 0) {
		/* timer 5, timebase 3 (1 s), uses own timer, 
		   not periodic, not reset, BOOT_TIMEOUT seconds,
		   calls back bootTimeout */
		timerWait(5, 3, FALSE, FALSE, FALSE, BOOT_TIMEOUT, cb);
	}
}

/** \brief stop usb timer
 * \note called from <b>Interrupt context</b> */
void utilStopBootTimeout() {
	timerStop(5);
}

/** \brief start flashing blue led (200 ms off, 50 ms on)
 * \note uses timers 8 and 9
 */
void utilFlashLedStart() {
	/* timer 8, timebase 2 (1ms), own timer, periodic, reset, 250 ms */ 
	timerWait(8, 2, FALSE, TRUE, TRUE, 300, &flashLed);
	/* timer 9, timebase 2 (1ms), matched agains timer 8, periodic, not reset,
		200 ms */
	timerWait(9, 2, TRUE, TRUE, FALSE, 200, &flashLed);
}

/** \brief flash timer timeout occured
 *
 * flash blue led
 * \param timerId timer which raised the event (8 or 9) */
void flashLed(uint8_t timerId) {
	if (timerId == 8) {
		/* turn led off */
		ledOff(LED_BLUE);
	} else {
		/* turn led on */
		ledOn(LED_BLUE);
	}
}

/** \brief flash timer stop */
void utilFlashLedStop() {
	/* stop both timers */
	timerStop(8);
	timerStop(9);
}

